﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Threading;

using Newtonsoft.Json;

using eSkin4Common;

namespace eSkin4Bitalino
{
    public class BitalinoHandler
    {
        public const string TopicSourceTag = "[Source]";
        public const string TopicNameTag = "[Name]";
        public const string BitalinoTopic1 = "/[Name]/Frame";

        private const int BitalinoNumberReadFrames = 1;

        private bool bitalinoIsConnected = false;
        private int sensorIndex = -1;
        private string bitalinoAddress = null;
        private string bitalinoName = null;
        private Bitalino bitalinoDevice;
        private Bitalino.Frame[] bitalinoFrames;
        private Thread bitalinoAquisitionThread;
        private  bool bitalinoDataAquisitionOn = false;
        private int bitalinoSampleRate = 1;

        private MqttHandler mqttHandler = null;

        private OscHandler oscHandler = null;

        private Dispatcher uiDispatcher = null;
        private List<Label> uiLabels = null;

        /// <summary>
        /// Create a bitalino handler for specified index or address.
        /// </summary>
        /// <param name="sensorIndex">Index number (0, 1, or 2)</param>
        /// <param name="bitalinoAddress">Address of bitalino device (if known); if null, auto-discover sensor with given index.</param>
        /// <param name="bitalinoName">Name of the device for messaging.</param>
        /// <param name="bitalinoSampleRate">Sample rate of bitalino; Supported values are 1, 10, or 100.</param>
        /// <param name="mqttHandler">MQTT handler to send messages to.</param>
        /// <param name="oscHandler">OSC handler to send messages to.</param>
        /// <param name="uiDispatcher">UI dispatcher for draw requests.</param>
        /// <param name="uiLabels">UI labels to populate with the received data.</param>
        public BitalinoHandler(
            int sensorIndex,
            // Device
            string bitalinoAddress,
            string bitalinoName,
            int bitalinoSampleRate, 
            // Messaging
            MqttHandler mqttHandler, 
            OscHandler oscHandler,
            // UI
            Dispatcher uiDispatcher, 
            List<Label> uiLabels)
        {
            this.sensorIndex = sensorIndex;
            this.bitalinoAddress = bitalinoAddress;
            this.bitalinoName = bitalinoName;
            this.bitalinoSampleRate = bitalinoSampleRate;
            this.mqttHandler = mqttHandler;
            this.oscHandler = oscHandler;
            this.uiDispatcher = uiDispatcher;
            this.uiLabels = uiLabels;
        }

        /// <summary>
        /// Get the connected state.
        /// </summary>
        public bool IsConnected
        {
            get { return bitalinoIsConnected; }
        }

        /// <summary>
        /// Set new bitalino address
        /// </summary>
        public string BitalinoAddress
        {
            set { bitalinoAddress = value; }
        }

        #region Bitalino handlers

        /// <summary>
        /// Connect BITalino sensor
        /// </summary>
        public bool OpenBitalino()
        {
            try
            {
                Console.WriteLine("Bitalino {0}: Opening connection to sensor.", sensorIndex);
                bitalinoIsConnected = false;
                if (string.IsNullOrEmpty(bitalinoAddress))
                {
                    Console.WriteLine("Bitalino {0}: Detecting devices ...", sensorIndex);
                    Bitalino.DevInfo[] bitalinoDevices = Bitalino.find();
                    if (sensorIndex >= 0 && bitalinoDevices != null && bitalinoDevices.Length < sensorIndex)
                    {
                        bitalinoAddress = bitalinoDevices[sensorIndex].macAddr.ToString();
                        Console.WriteLine("Bitalino {0}: Using address {1} ...", sensorIndex, bitalinoAddress);
                    }
                }

                if (string.IsNullOrEmpty(bitalinoAddress))
                {
                    Console.WriteLine("Bitalino {0}: No device found ...", sensorIndex);
                }
                else
                {
                    Console.WriteLine("Bitalino {0}: Connecting to address {1} ...", sensorIndex, bitalinoAddress);
                    bitalinoDevice = new Bitalino(bitalinoAddress);

                    // Get device version string
                    string bitalinoVersion = bitalinoDevice.version();
                    string bitalinoStatusText = string.Format("Bitalino {0}: Address {1} version {2} is now connected.", sensorIndex, bitalinoAddress, bitalinoVersion);
                    Console.WriteLine(bitalinoStatusText);

                    // Start reading data
                    bitalinoDevice.start(bitalinoSampleRate, new int[] { 0, 1, 2, 3, 4, 5 }); // start acquisition of all analog channels

                    // Initialize bitalino data frames
                    bitalinoFrames = new Bitalino.Frame[BitalinoNumberReadFrames];
                    for (int i = 0; i < bitalinoFrames.Length; i++)
                    {
                        bitalinoFrames[i] = new Bitalino.Frame();
                    }

                    // Start data aquisition thread
                    bitalinoAquisitionThread = new Thread(new ThreadStart(BitalinoDataAquisition));
                    bitalinoAquisitionThread.Start();

                    bitalinoStatusText = string.Format("Bitalino {0}: data aquisition at {1} Hz has been started.", sensorIndex, bitalinoSampleRate);
                    Console.WriteLine(bitalinoStatusText);

                    // Mark as connected
                    bitalinoIsConnected = true;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("Bitalino {0}: error opening sensor.", sensorIndex);
                Console.WriteLine(e.Message);
            }

            return bitalinoIsConnected;
        }

        /// <summary>
        /// Disconnect BITalino sensor
        /// </summary>
        public void CloseBitalino()
        {
            try
            {
                // Mark as disconnected
                bitalinoIsConnected = false;

                // Stop data aquisition thread
                if (bitalinoAquisitionThread != null)
                {
                    Console.WriteLine("Bitalino {0}: Stopping data aquisition thread.", sensorIndex);
                    bitalinoDataAquisitionOn = false;
                    bitalinoAquisitionThread.Join(TimeSpan.FromSeconds(1));
                    bitalinoAquisitionThread.Abort();
                    bitalinoAquisitionThread = null;
                }

                // Close connection
                if (bitalinoDevice != null)
                {
                    Console.WriteLine("Bitalino {0}: Closing connection to sensor.", sensorIndex);
                    bitalinoDevice.stop();
                    bitalinoDevice.Dispose();
                    bitalinoDevice = null;
                }
            }
            catch
            {
                Console.WriteLine("Bitalino {0}: Some error occured during closing. Ignoring.", sensorIndex);
            }
        }

        #region Bitalino Thread Handler

        private void BitalinoDataAquisition()
        {
            Console.WriteLine("Bitalino {0}: Data aquisition thread started ...", sensorIndex);

            bitalinoDataAquisitionOn = true;
            while (bitalinoDataAquisitionOn)
            {
                try
                {
                    // Get frames from device
                    bitalinoDevice.read(bitalinoFrames);
                }
                catch
                {
                    Console.WriteLine("Bitalino {0}: Error during data aquisition.", sensorIndex);
                    bitalinoDataAquisitionOn = false;
                }

                if (bitalinoDataAquisitionOn)
                {
                    // Use first frame in set
                    Bitalino.Frame f = bitalinoFrames[0];

                    // Update UI
                    uiDispatcher.Invoke((Action)(() =>
                    {
                        uiLabels[0].Content = this.bitalinoName;
                        uiLabels[1].Content = f.digital[0].ToString();
                        uiLabels[2].Content = f.digital[1].ToString();
                        uiLabels[3].Content = f.digital[2].ToString();
                        uiLabels[4].Content = f.digital[3].ToString();
                        uiLabels[5].Content = f.analog[0].ToString();
                        uiLabels[6].Content = f.analog[1].ToString();
                        uiLabels[7].Content = f.analog[2].ToString();
                        uiLabels[8].Content = f.analog[3].ToString();
                        uiLabels[9].Content = f.analog[4].ToString();
                        uiLabels[10].Content = f.analog[5].ToString();
                    }));

                    // Send data
                    foreach (Bitalino.Frame frame in bitalinoFrames)
                    {
                        // Send data
                        BitalinoFrameMessage bitalinoData = new BitalinoFrameMessage(bitalinoAddress, frame);
                        string bitalinoFrameMessageString = JsonConvert.SerializeObject(bitalinoData);
                        string topic = BitalinoTopic1.Replace(TopicSourceTag, bitalinoData.Source).Replace(TopicNameTag, this.bitalinoName);
                        if (this.mqttHandler != null && this.mqttHandler.IsConnected)
                        {
                            this.mqttHandler.SendMessage(topic, bitalinoFrameMessageString);
                        }
                        if (this.oscHandler != null && this.oscHandler.IsConnected)
                        {
                            this.oscHandler.SendMessage(topic, bitalinoFrameMessageString);
                        }
                    }
                }
            }

            Console.WriteLine("Bitalino {0}: Data aquisition thread stopped ...", sensorIndex);
        }

        #endregion

        #endregion
    }
}
