﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Threading;

// https://github.com/eclipse/paho.mqtt.m2mqtt
using uPLibrary.Networking.M2Mqtt;
using uPLibrary.Networking.M2Mqtt.Messages;

using System.Diagnostics;

namespace eSkin4Common
{
    public class MqttHandler : IDisposable
    {
        private string mqttDataSourceName = string.Empty;
        private bool mqttIsConnected = false;
        private string mqttBrokerHostname = null;
        private MqttClient mqttClient;
        private string mqttClientId;

        private bool saveMessages = false;
        private StreamWriter saveStream = null;

        private Dispatcher uiDispatcher = null;
        private List<Label> uiLabels = null;

        private long messageCount = 0;
        private long lastMessageCount = 0;
        private Stopwatch rateStopwatch = null;
        private DispatcherTimer rateDispatcherTimer = null;

        private bool disposed = false;

        /// <summary>
        /// Creates a new MqttHandler to a specified broker.
        /// </summary>
        /// <param name="dataSourceName">Name of the data source</param>
        /// <param name="brokerHostname">Broker hostname to connect to.</param>
        /// <param name="uiDispatcher">UI dispatcher for draw requests.</param>
        /// <param name="uiLabels">UI labels to populate with the received data.</param>
        public MqttHandler(string dataSourceName, string brokerHostname, Dispatcher uiDispatcher, List<Label> uiLabels)
        {
            this.mqttDataSourceName = dataSourceName;
            this.mqttBrokerHostname = brokerHostname;
            this.uiDispatcher = uiDispatcher;
            this.uiLabels = uiLabels;

            //  DispatcherTimer setup
            rateStopwatch = new Stopwatch();
            rateDispatcherTimer = new System.Windows.Threading.DispatcherTimer();
            rateDispatcherTimer.Tick += new EventHandler(RateDispatcherTimerTick);
            rateDispatcherTimer.Interval = new TimeSpan(0, 0, 1);
            rateDispatcherTimer.Start();
        }

        // Public implementation of Dispose pattern callable by consumers.
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        // Protected implementation of Dispose pattern.
        protected virtual void Dispose(bool disposing)
        {
            if (disposed)
            {
                return;
            }

            if (disposing)
            {
                // Stop timers
                rateDispatcherTimer.Stop();

                // Close log file
                if (saveStream != null)
                {
                    saveStream.Close();
                    saveStream = null;
                }
            }

            disposed = true;
        }

        /// <summary>
        /// Save messages to a file.
        /// </summary>
        public bool SaveMessages
        {
            set
            {
                if (this.saveMessages != value)
                {
                    this.saveMessages = value;
                    if (this.saveMessages)
                    {
                        // open
                        string filename = string.Format("{0}-Messages-{1}.txt", mqttDataSourceName, DateTime.Now.ToString("yyyyMMdd_HHmmss"));
                        saveStream = File.AppendText(filename);

                        Console.WriteLine("Saving messages to {0}.", filename);
                    }
                    else
                    {
                        // close
                        saveStream.Close();
                        saveStream = null;

                        Console.WriteLine("Message save ended.");
                    }
                }
            }
        }

        /// <summary>
        /// Get the connected state.
        /// </summary>
        public bool IsConnected {
            get { return mqttIsConnected; }
        }

        /// <summary>
        /// Set new broker hostname.
        /// </summary>
        public string BrokerHostname {
            set { mqttBrokerHostname = value; }
        }

        #region Mqtt handlers

        /// <summary>
        /// Connect to MQTT broker
        /// </summary>
        public bool OpenMqtt()
        {
            try
            {
                Console.WriteLine("Connecting to MQTT broker: {0}", mqttBrokerHostname);
                mqttClient = new MqttClient(mqttBrokerHostname);

                mqttClientId = Guid.NewGuid().ToString();
                mqttClient.Connect(mqttClientId);

                Console.WriteLine("Connected to MQTT broker");
                mqttIsConnected = true;

                // Track runtime
                rateStopwatch.Reset();
                rateStopwatch.Start();
            }
            catch (Exception e)
            {
                Console.WriteLine("Error connecting to MQTT broker.");
                Console.WriteLine(e.Message);
                mqttIsConnected = false;

                // Track runtime
                rateStopwatch.Stop();
            }

            return mqttIsConnected;
        }

        /// <summary>
        /// Disconnect from MQTT broker
        /// </summary>
        public void CloseMqtt()
        {
            if (mqttClient != null && mqttClient.IsConnected)
            {
                Console.WriteLine("Disconnecting from MQTT broker ...");
                mqttClient.Disconnect();
                mqttClient = null;
                mqttIsConnected = false;

                // Track runtime
                rateStopwatch.Stop();
            }

            Console.WriteLine("Disconnected from MQTT broker");
        }

        public void SendMessage(string topic, string dataMessage)
        {
            // Count messages
            messageCount++;

            // Save message
            if (this.saveMessages && this.saveStream != null)
            {                
                saveStream.WriteLine(dataMessage);
            }

            // Send data
            if (mqttIsConnected)
            {
                mqttClient.Publish(topic, Encoding.UTF8.GetBytes(dataMessage), MqttMsgBase.QOS_LEVEL_EXACTLY_ONCE, false);
            }
        }

        //  System.Windows.Threading.DispatcherTimer.Tick handler
        //
        //  Updates the current seconds display and calls
        //  InvalidateRequerySuggested on the CommandManager to force 
        //  the Command to raise the CanExecuteChanged event.
        private void RateDispatcherTimerTick(object sender, EventArgs e)
        {
            // Update UI
            uiDispatcher.Invoke((Action)(() =>
            {
                // Calculate message rate
                double messageRate = 0.0;
                if (rateStopwatch.IsRunning)
                {
                    messageRate = Math.Round(1000.0 * (double)(messageCount - lastMessageCount) / (double)rateStopwatch.ElapsedMilliseconds, 1);
                }

                // Display message count and rate
                uiLabels[0].Content = string.Format("{0} ({1}/s)", messageCount.ToString("N0"), messageRate);

                // Reset stopwatch and capture current message count
                rateStopwatch.Reset();
                rateStopwatch.Start();
                lastMessageCount = messageCount;
            }));
        }

        #endregion
    }
}
