﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Threading;

// https://bitbucket.org/rugcode/rug.osc/
using Rug.Osc;
using System.Net;

namespace eSkin4Common
{
    public class OscHandler : IDisposable
    {
        private bool oscIsConnected = false;
        private string oscAddress = null;
        private string oscPort = null;
        private OscSender oscSender;

        private Dispatcher uiDispatcher = null;
        private List<Label> uiLabels = null;

        private bool disposed = false;

        /// <summary>
        /// Creates a new OscHandler to a specified IP and port.
        /// </summary>
        /// <param name="oscAddress">IP address or hostname of the OSC client</param>
        /// <param name="oscPort">Port of the OSC client.</param>
        /// <param name="uiDispatcher">UI dispatcher for draw requests.</param>
        /// <param name="uiLabels">UI labels to populate with the received data.</param>  
        public OscHandler(string oscAddress, string oscPort, Dispatcher uiDispatcher, List<Label> uiLabels)
        {
            this.oscAddress = oscAddress;
            this.oscPort = oscPort;
            this.uiDispatcher = uiDispatcher;
            this.uiLabels = uiLabels;
        }

        // Public implementation of Dispose pattern callable by consumers.
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        // Protected implementation of Dispose pattern.
        protected virtual void Dispose(bool disposing)
        {
            if (disposed)
            {
                return;
            }

            if (disposing)
            {
                if (oscSender != null)
                {
                    oscSender.Close();
                    oscSender.Dispose();
                    oscSender = null;
                }
            }

            disposed = true;
        }

        /// <summary>
        /// Get the connected state.
        /// </summary>
        public bool IsConnected
        {
            get { return oscIsConnected; }
        }

        /// <summary>
        /// Set new address.
        /// </summary>
        public string Address
        {
            set { oscAddress = value; }
        }

        /// <summary>
        /// Set new port.
        /// </summary>
        public string Port
        {
            set { oscPort = value; }
        }

        #region Osc handlers

        /// <summary>
        /// Open OSC sender
        /// </summary>
        public bool OpenOsc()
        {
            try
            {
                Console.WriteLine("Opening OSC sender: {0} port {1}", oscAddress, oscPort);
                IPAddress address = IPAddress.Parse(oscAddress);
                int port = int.Parse(oscPort);
                oscSender = new OscSender(address, port);
                Console.WriteLine("OSC sender created.");
                oscSender.Connect();
                Console.WriteLine("OSC sender connected.");

                oscIsConnected = true;
            }
            catch (Exception e)
            {
                Console.WriteLine("Error creating OSC sender.");
                Console.WriteLine(e.Message);
                oscIsConnected = false;
                oscSender = null;
            }

            return oscIsConnected;
        }

        /// <summary>
        /// Close OSC sender
        /// </summary>
        public void CloseOsc()
        {
            if (oscSender != null)
            {
                Console.WriteLine("Sending remaining messages ...");
                oscSender.WaitForAllMessagesToComplete();
                oscSender.Close();
                Console.WriteLine("OSC disconnected.");
                oscSender.Dispose();
                oscSender = null;
                oscIsConnected = false;
            }

            Console.WriteLine("Closed OSC sender.");
        }

        public void SendMessage(string topic, string dataMessage)
        {
            // Send data
            if (oscIsConnected)
            {
                OscMessage oscMessage = new OscMessage(topic, dataMessage);
                oscSender.Send(oscMessage);
            }
        }

        #endregion
    }
}
