﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using Microsoft.Kinect;

// https://github.com/eclipse/paho.mqtt.m2mqtt
using uPLibrary.Networking.M2Mqtt;
using uPLibrary.Networking.M2Mqtt.Messages;
using System.Net;

using Newtonsoft.Json;

using eSkin4Common;

namespace eSkin4Kinect
{

    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        public const string TopicJointTag = "[Joint]";
        private static string KinectTopic1 = "/Kinect/[Joint]";

        private const string BodyIdFormat = "x8";
        private const string NumberDisplayFormat = "0.000";

        private static bool mqttIsEnabled = bool.Parse(ConfigurationManager.AppSettings["mqttIsEnabled"]);
        MqttHandler mqttHandler = null;

        private static bool oscIsEnabled = bool.Parse(ConfigurationManager.AppSettings["oscIsEnabled"]);
        string oscAddressSetting = ConfigurationManager.AppSettings["oscAddress"];
        string oscPortSetting = ConfigurationManager.AppSettings["oscPort"];
        OscHandler oscHandler = null;

        private static bool kinectIsEnabled = bool.Parse(ConfigurationManager.AppSettings["kinectIsEnabled"]);
        private static string kinectJointsSetting = ConfigurationManager.AppSettings["kinectJoints"];
        List<JointType> kinectJoints = new List<JointType>();
        private bool kinectIsConnected = false;
        KinectSensor kinectSensor;
        MultiSourceFrameReader kinectReader;
        IList<Body> kinectBodies;

        /// <summary>
        /// Main program window .ctor
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();

            // Connect to console
            TextBoxConsole.RedirectConsoleToTextBox(messageConsole);
            Console.WriteLine("eSkin4Kinect started");

            // Get default configuration and create handlers
            if (mqttIsEnabled)
            {
                string mqttBrokerHostname = ConfigurationManager.AppSettings["mqttBrokerHostname"];
                if (!string.IsNullOrEmpty(mqttBrokerHostname))
                {
                    mqttBroker.Text = mqttBrokerHostname;
                }

                List<Label> mqttLabels = new List<Label>();
                mqttLabels.Add(this.mqttMessages);

                this.mqttHandler = new MqttHandler("Kinect", mqttBrokerHostname, this.Dispatcher, mqttLabels);
            }

            if (oscIsEnabled)
            {
                if (!string.IsNullOrEmpty(oscAddressSetting))
                {
                    oscAddress.Text = oscAddressSetting;
                }

                if (!string.IsNullOrEmpty(oscPortSetting))
                {
                    oscPort.Text = oscPortSetting;
                }

                List<Label> oscLabels = new List<Label>();
                oscHandler = new OscHandler(oscAddressSetting, oscPortSetting, this.Dispatcher, oscLabels);
            }

            // Prepare kinect
            if (kinectIsEnabled)
            {
                if ("ALL".Equals(kinectJointsSetting.ToUpperInvariant()))
                {
                    // Use all joints
                    kinectJoints = Enum.GetValues(typeof(JointType)).Cast<JointType>().ToList();
                }
                else
                {
                    // Parse active joints
                    kinectJoints = kinectJointsSetting.Split(',').Select(s => (JointType)Enum.Parse(typeof(JointType), s)).ToList();
                }
            }
        }

        #region Window handlers

        /// <summary>
        /// Startup program
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            if (mqttIsEnabled)
            {
                // Open MQTT 
                bool mqttIsConnected = mqttHandler.OpenMqtt();
                mqttStatusIndicator.Fill = mqttIsConnected ? Brushes.Green : Brushes.Red;

                // Set topic in UI
                this.mqttTopic1.Content = KinectTopic1;
            }

            if (oscIsEnabled)
            {
                bool oscIsConnected = oscHandler.OpenOsc();
                oscStatusIndicator.Fill = oscIsConnected ? Brushes.Green : Brushes.Red;
            }

            if (kinectIsEnabled)
            {
                // Open Kinect
                kinectIsConnected = OpenKinect();
                kinectStatusIndicator.Fill = kinectIsConnected ? Brushes.Green : Brushes.Red;

                Console.WriteLine("Messaging positions of joints: {0}", string.Join(", ", kinectJoints));
            }
        }

        /// <summary>
        /// Shutdown program
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Window_Closed(object sender, EventArgs e)
        {
            if (mqttIsEnabled)
            {
                // Close MQTT 
                mqttHandler.CloseMqtt();
                mqttStatusIndicator.Fill = Brushes.LightGray;
            }

            if (oscIsEnabled)
            {
                // Close OSC
                oscHandler.CloseOsc();
                oscStatusIndicator.Fill = Brushes.LightGray;
            }

            if (kinectIsEnabled)
            {
                // Close Kinect
                CloseKinect();
                kinectIsConnected = false;
                kinectStatusIndicator.Fill = Brushes.LightGray;
            }
        }

        #endregion

        #region Kinect handlers

        /// <summary>
        /// Connect Kinect
        /// </summary>
        private bool OpenKinect()
        {
            try
            {
                Console.WriteLine("Opening connection to Kinect sensor.");

                // Get default sensor
                kinectSensor = KinectSensor.GetDefault();
                if (kinectSensor != null)
                {
                    // Connect to sensor
                    kinectSensor.Open();

                    // Open data reader
                    kinectReader = kinectSensor.OpenMultiSourceFrameReader(FrameSourceTypes.Color | FrameSourceTypes.Body);

                    // Hook up event handler
                    kinectReader.MultiSourceFrameArrived += KinectReader_MultiSourceFrameArrived;
                }
                else
                {
                    Console.WriteLine("No Kinect sensor found.");
                    return false;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("Error opening Kinect sensor.");
                Console.WriteLine(e.Message);
                return false;
            }

            return true;
        }

        /// <summary>
        /// Disconnect Kinect
        /// </summary>
        private void CloseKinect()
        {
            if (kinectReader != null)
            {
                Console.WriteLine("Stopping Kinect data stream.");
                kinectReader.Dispose();
            }

            if (kinectSensor != null)
            {
                Console.WriteLine("Closing connection to Kinect sensor.");
                kinectSensor.Close();
            }
        }

        #region Kinect Event Handlers

        /// <summary>
        /// Handle Kinect data
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void KinectReader_MultiSourceFrameArrived(object sender, MultiSourceFrameArrivedEventArgs e)
        {
            var reference = e.FrameReference.AcquireFrame();

            // Color channel
            using (var frame = reference.ColorFrameReference.AcquireFrame())
            {
                if (frame != null)
                {
                    kinectCamera.Source = ColorFrameToBitmap(frame);
                }
            }

            // Body data
            using (var frame = reference.BodyFrameReference.AcquireFrame())
            {
                if (frame != null)
                {
                    kinectCanvas.Children.Clear();

                    kinectBodies = new Body[frame.BodyFrameSource.BodyCount];

                    frame.GetAndRefreshBodyData(kinectBodies);

                    int numTrackedBody = 0;
                    foreach (var body in kinectBodies)
                    {
                        if (body != null)
                        {
                            if (body.IsTracked)
                            {
                                // Enumerate 
                                numTrackedBody++;

                                // Draw skeleton.
                                DrawSkeleton(kinectCanvas, body);

                                // Get body id (as Hex)
                                string bodyId = (body.TrackingId & (ulong)0x00000000ffffffff).ToString(BodyIdFormat);

                                // Show head joint as sample
                                JointType sampleJointType = JointType.Head;

                                // Update UI for tracked bodies
                                this.Dispatcher.Invoke((Action)(() =>
                                {
                                    if (numTrackedBody == 1)
                                    {
                                        this.kinectHead1BodyId.Content = bodyId;
                                        this.kinectHead1X.Content = body.Joints[sampleJointType].Position.X.ToString(NumberDisplayFormat);
                                        this.kinectHead1Y.Content = body.Joints[sampleJointType].Position.Y.ToString(NumberDisplayFormat);
                                        this.kinectHead1Z.Content = body.Joints[sampleJointType].Position.Z.ToString(NumberDisplayFormat);
                                        this.kinectHead1StatusIndicator.Fill = Brushes.Green;

                                    }
                                    else
                                    if (numTrackedBody == 2)
                                    {
                                        this.kinectHead2BodyId.Content = bodyId;
                                        this.kinectHead2X.Content = body.Joints[sampleJointType].Position.X.ToString(NumberDisplayFormat);
                                        this.kinectHead2Y.Content = body.Joints[sampleJointType].Position.Y.ToString(NumberDisplayFormat);
                                        this.kinectHead2Z.Content = body.Joints[sampleJointType].Position.Z.ToString(NumberDisplayFormat);
                                        this.kinectHead2StatusIndicator.Fill = Brushes.Green;
                                    }
                                    else
                                    if (numTrackedBody == 3)
                                    {
                                        this.kinectHead3BodyId.Content = bodyId;
                                        this.kinectHead3X.Content = body.Joints[sampleJointType].Position.X.ToString(NumberDisplayFormat);
                                        this.kinectHead3Y.Content = body.Joints[sampleJointType].Position.Y.ToString(NumberDisplayFormat);
                                        this.kinectHead3Z.Content = body.Joints[sampleJointType].Position.Z.ToString(NumberDisplayFormat);
                                        this.kinectHead3StatusIndicator.Fill = Brushes.Green;
                                    }
                                    else
                                    if (numTrackedBody == 4)
                                    {
                                        this.kinectHead4BodyId.Content = bodyId;
                                        this.kinectHead4X.Content = body.Joints[sampleJointType].Position.X.ToString(NumberDisplayFormat);
                                        this.kinectHead4Y.Content = body.Joints[sampleJointType].Position.Y.ToString(NumberDisplayFormat);
                                        this.kinectHead4Z.Content = body.Joints[sampleJointType].Position.Z.ToString(NumberDisplayFormat);
                                        this.kinectHead4StatusIndicator.Fill = Brushes.Green;
                                    }
                                }));

                                // Send data
                                foreach (JointType jointType in kinectJoints)
                                {
                                    KinectBodyMessage kinectData = new KinectBodyMessage();
                                    kinectData.BodyId = bodyId;
                                    kinectData.JointId = jointType.ToString();
                                    kinectData.X = Math.Round((double)body.Joints[jointType].Position.X, 3);
                                    kinectData.Y = Math.Round((double)body.Joints[jointType].Position.Y, 3);
                                    kinectData.Z = Math.Round((double)body.Joints[jointType].Position.Z, 3);
                                    string kinectDataMessage = JsonConvert.SerializeObject(kinectData);
                                    string topic = KinectTopic1.Replace(TopicJointTag, kinectData.JointId);

                                    if (this.mqttHandler != null && this.mqttHandler.IsConnected)
                                    {
                                        this.mqttHandler.SendMessage(topic, kinectDataMessage);
                                    }
                                    if (this.oscHandler != null && this.oscHandler.IsConnected)
                                    {
                                        this.oscHandler.SendMessage(topic, kinectDataMessage);
                                    }
                                }
                            }
                        }
                    }

                    // Update UI for untracked bodies
                    this.Dispatcher.Invoke((Action)(() =>
                    {
                        if (numTrackedBody <= 0)
                        {
                            this.kinectHead1BodyId.Content = string.Empty;
                            this.kinectHead1X.Content = string.Empty;
                            this.kinectHead1Y.Content = string.Empty;
                            this.kinectHead1Z.Content = string.Empty;
                            this.kinectHead1StatusIndicator.Fill = Brushes.LightGray;
                        }
                        if (numTrackedBody <= 1)
                        {
                            this.kinectHead2BodyId.Content = string.Empty;
                            this.kinectHead2X.Content = string.Empty;
                            this.kinectHead2Y.Content = string.Empty;
                            this.kinectHead2Z.Content = string.Empty;
                            this.kinectHead2StatusIndicator.Fill = Brushes.LightGray;
                        }
                        if (numTrackedBody <= 2)
                        {
                            this.kinectHead3BodyId.Content = string.Empty;
                            this.kinectHead3X.Content = string.Empty;
                            this.kinectHead3Y.Content = string.Empty;
                            this.kinectHead3Z.Content = string.Empty;
                            this.kinectHead3StatusIndicator.Fill = Brushes.LightGray;
                        }
                        if (numTrackedBody <= 3)
                        {
                            this.kinectHead4BodyId.Content = string.Empty;
                            this.kinectHead4X.Content = string.Empty;
                            this.kinectHead4Y.Content = string.Empty;
                            this.kinectHead4Z.Content = string.Empty;
                            this.kinectHead4StatusIndicator.Fill = Brushes.LightGray;
                        }
                    }));
                }
            }
        }

        #endregion

        #region Private Kinect helpers

        private static ImageSource ColorFrameToBitmap(ColorFrame frame)
        {
            int width = frame.FrameDescription.Width;
            int height = frame.FrameDescription.Height;
            PixelFormat format = PixelFormats.Bgr32;

            byte[] pixels = new byte[width * height * ((format.BitsPerPixel + 7) / 8)];

            if (frame.RawColorImageFormat == ColorImageFormat.Bgra)
            {
                frame.CopyRawFrameDataToArray(pixels);
            }
            else
            {
                frame.CopyConvertedFrameDataToArray(pixels, ColorImageFormat.Bgra);
            }

            int stride = width * format.BitsPerPixel / 8;

            return BitmapSource.Create(width, height, 96, 96, format, null, pixels, stride);
        }


        private static Joint ScaleTo(Joint joint, double width, double height, float skeletonMaxX, float skeletonMaxY)
        {
            joint.Position = new CameraSpacePoint
            {
                X = Scale(width, skeletonMaxX, joint.Position.X),
                Y = Scale(height, skeletonMaxY, -joint.Position.Y),
                Z = joint.Position.Z
            };

            return joint;
        }

        private static Joint ScaleTo(Joint joint, double width, double height)
        {
            return ScaleTo(joint, width, height, 1.0f, 1.0f);
        }

        private static float Scale(double maxPixel, double maxSkeleton, float position)
        {
            float value = (float)((((maxPixel / maxSkeleton) / 2) * position) + (maxPixel / 2));

            if (value > maxPixel)
            {
                return (float)maxPixel;
            }

            if (value < 0)
            {
                return 0;
            }

            return value;
        }

        private static void DrawSkeleton(Canvas canvas, Body body)
        {
            if (body == null) return;

            foreach (Joint joint in body.Joints.Values)
            {
                DrawPoint(canvas, joint);
            }

            DrawLine(canvas, body.Joints[JointType.Head], body.Joints[JointType.Neck]);
            DrawLine(canvas, body.Joints[JointType.Neck], body.Joints[JointType.SpineShoulder]);
            DrawLine(canvas, body.Joints[JointType.SpineShoulder], body.Joints[JointType.ShoulderLeft]);
            DrawLine(canvas, body.Joints[JointType.SpineShoulder], body.Joints[JointType.ShoulderRight]);
            DrawLine(canvas, body.Joints[JointType.SpineShoulder], body.Joints[JointType.SpineMid]);
            DrawLine(canvas, body.Joints[JointType.ShoulderLeft], body.Joints[JointType.ElbowLeft]);
            DrawLine(canvas, body.Joints[JointType.ShoulderRight], body.Joints[JointType.ElbowRight]);
            DrawLine(canvas, body.Joints[JointType.ElbowLeft], body.Joints[JointType.WristLeft]);
            DrawLine(canvas, body.Joints[JointType.ElbowRight], body.Joints[JointType.WristRight]);
            DrawLine(canvas, body.Joints[JointType.WristLeft], body.Joints[JointType.HandLeft]);
            DrawLine(canvas, body.Joints[JointType.WristRight], body.Joints[JointType.HandRight]);
            DrawLine(canvas, body.Joints[JointType.HandLeft], body.Joints[JointType.HandTipLeft]);
            DrawLine(canvas, body.Joints[JointType.HandRight], body.Joints[JointType.HandTipRight]);
            DrawLine(canvas, body.Joints[JointType.HandTipLeft], body.Joints[JointType.ThumbLeft]);
            DrawLine(canvas, body.Joints[JointType.HandTipRight], body.Joints[JointType.ThumbRight]);
            DrawLine(canvas, body.Joints[JointType.SpineMid], body.Joints[JointType.SpineBase]);
            DrawLine(canvas, body.Joints[JointType.SpineBase], body.Joints[JointType.HipLeft]);
            DrawLine(canvas, body.Joints[JointType.SpineBase], body.Joints[JointType.HipRight]);
            DrawLine(canvas, body.Joints[JointType.HipLeft], body.Joints[JointType.KneeLeft]);
            DrawLine(canvas, body.Joints[JointType.HipRight], body.Joints[JointType.KneeRight]);
            DrawLine(canvas, body.Joints[JointType.KneeLeft], body.Joints[JointType.AnkleLeft]);
            DrawLine(canvas, body.Joints[JointType.KneeRight], body.Joints[JointType.AnkleRight]);
            DrawLine(canvas, body.Joints[JointType.AnkleLeft], body.Joints[JointType.FootLeft]);
            DrawLine(canvas, body.Joints[JointType.AnkleRight], body.Joints[JointType.FootRight]);
        }

        public static void DrawPoint(Canvas canvas, Joint joint)
        {
            if (joint.TrackingState == TrackingState.NotTracked) return;

            joint = ScaleTo(joint, canvas.ActualWidth, canvas.ActualHeight);

            Ellipse ellipse = new Ellipse
            {
                Width = 12,
                Height = 12,
                Fill = new SolidColorBrush(Colors.LightBlue)
            };

            Canvas.SetLeft(ellipse, joint.Position.X - ellipse.Width / 2);
            Canvas.SetTop(ellipse, joint.Position.Y - ellipse.Height / 2);

            canvas.Children.Add(ellipse);
        }

        private static void DrawLine(Canvas canvas, Joint first, Joint second)
        {
            if (first.TrackingState == TrackingState.NotTracked || second.TrackingState == TrackingState.NotTracked) return;

            first = ScaleTo(first, canvas.ActualWidth, canvas.ActualHeight);
            second = ScaleTo(second, canvas.ActualWidth, canvas.ActualHeight);

            Line line = new Line
            {
                X1 = first.Position.X,
                Y1 = first.Position.Y,
                X2 = second.Position.X,
                Y2 = second.Position.Y,
                StrokeThickness = 6,
                Stroke = new SolidColorBrush(Colors.LightBlue)
            };

            canvas.Children.Add(line);
        }

        #endregion

        #endregion

        private void mqttReconnect_Click(object sender, RoutedEventArgs e)
        {
            if (mqttIsEnabled)
            {
                // Close MQTT 
                mqttHandler.CloseMqtt();
                mqttStatusIndicator.Fill = Brushes.LightGray;

                // Update hostname from UI
                mqttHandler.BrokerHostname = mqttBroker.Text;

                // Open MQTT 
                bool mqttIsConnected = mqttHandler.OpenMqtt();
                mqttStatusIndicator.Fill = mqttIsConnected ? Brushes.Green : Brushes.Red;
            }
        }

        private void oscReconnect_Click(object sender, RoutedEventArgs e)
        {
            if (oscIsEnabled)
            {
                // Close OSC
                oscHandler.CloseOsc();
                oscStatusIndicator.Fill = Brushes.LightGray;

                // Update address and port from UI
                oscHandler.Address = oscAddress.Text;
                oscHandler.Port = oscPort.Text;

                // Open OSC
                bool oscIsConnected = oscHandler.OpenOsc();
                oscStatusIndicator.Fill = oscIsConnected ? Brushes.Green : Brushes.Red;
            }
        }

        private void CheckBox_Checked(object sender, RoutedEventArgs e)
        {
            mqttHandler.SaveMessages = true;
        }

        private void CheckBox_Unchecked(object sender, RoutedEventArgs e)
        {
            mqttHandler.SaveMessages = false;
        }

        /// <summary>
        /// Configuration button click: opens Notepad to edit App.config
        /// </summary>
        private void Button_Click(object sender, RoutedEventArgs e)
        {
            ConfigurationFile.Edit();
        }
    }
}
