﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using MyoSharp.Communication;
using MyoSharp.Device;
using MyoSharp.Exceptions;

// https://github.com/eclipse/paho.mqtt.m2mqtt
using uPLibrary.Networking.M2Mqtt;
using uPLibrary.Networking.M2Mqtt.Messages;
using System.Net;

using Newtonsoft.Json;

using eSkin4Common;

namespace eSkin4Myo
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private static bool myoIsEnabled = bool.Parse(ConfigurationManager.AppSettings["myoIsEnabled"]);
        MyoHandler myoHandler = null;
        private static int myoRateReductionLevel = int.Parse(ConfigurationManager.AppSettings["myoRateReductionLevel"]);

        private bool gyroscopeEnabled = bool.Parse(ConfigurationManager.AppSettings["gyroscopeEnabled"]);
        private bool accelerometerEnabled = bool.Parse(ConfigurationManager.AppSettings["accelerometerEnabled"]);
        private bool orientationEnabled = bool.Parse(ConfigurationManager.AppSettings["orientationEnabled"]);

        private static bool mqttIsEnabled = bool.Parse(ConfigurationManager.AppSettings["mqttIsEnabled"]);
        MqttHandler mqttHandler = null;

        private static bool oscIsEnabled = bool.Parse(ConfigurationManager.AppSettings["oscIsEnabled"]);
        string oscAddressSetting = ConfigurationManager.AppSettings["oscAddress"];
        string oscPortSetting = ConfigurationManager.AppSettings["oscPort"];
        OscHandler oscHandler = null;

        /// <summary>
        /// Main program window .ctor
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();

            // Connect to console
            TextBoxConsole.RedirectConsoleToTextBox(messageConsole);
            Console.WriteLine("eSkin4Myo started");

            // Get default configuration and create handlers
            if (mqttIsEnabled)
            {
                string mqttBrokerHostname = ConfigurationManager.AppSettings["mqttBrokerHostname"];
                if (!string.IsNullOrEmpty(mqttBrokerHostname))
                {
                    mqttBroker.Text = mqttBrokerHostname;
                }

                List<Label> mqttLabels = new List<Label>();
                mqttLabels.Add(this.mqttMessages);

                mqttHandler = new MqttHandler("Myo", mqttBrokerHostname, this.Dispatcher, mqttLabels);                
            }

            if (oscIsEnabled)
            {
                if (!string.IsNullOrEmpty(oscAddressSetting))
                {
                    oscAddress.Text = oscAddressSetting;
                }

                if (!string.IsNullOrEmpty(oscPortSetting))
                {
                    oscPort.Text = oscPortSetting;
                }


                List<Label> oscLabels = new List<Label>();
                oscHandler = new OscHandler(oscAddressSetting, oscPortSetting, this.Dispatcher, oscLabels);
            }

            if (myoIsEnabled)
            {
                // Initialize UI labels
                List<Label> myoLabels = new List<Label>();
                myoLabels.Add(this.myo1Address);
                myoLabels.Add(this.myo1Pose);
                myoLabels.Add(this.myo1GyroscopeX);
                myoLabels.Add(this.myo1GyroscopeY);
                myoLabels.Add(this.myo1GyroscopeZ);
                myoLabels.Add(this.myo1AccelerometerX);
                myoLabels.Add(this.myo1AccelerometerY);
                myoLabels.Add(this.myo1AccelerometerZ);
                myoLabels.Add(this.myo1OrientationW);
                myoLabels.Add(this.myo1OrientationX);
                myoLabels.Add(this.myo1OrientationY);
                myoLabels.Add(this.myo1OrientationZ);
                //
                myoLabels.Add(this.myo2Address);
                myoLabels.Add(this.myo2Pose);
                myoLabels.Add(this.myo2GyroscopeX);
                myoLabels.Add(this.myo2GyroscopeY);
                myoLabels.Add(this.myo2GyroscopeZ);
                myoLabels.Add(this.myo2AccelerometerX);
                myoLabels.Add(this.myo2AccelerometerY);
                myoLabels.Add(this.myo2AccelerometerZ);
                myoLabels.Add(this.myo2OrientationW);
                myoLabels.Add(this.myo2OrientationX);
                myoLabels.Add(this.myo2OrientationY);
                myoLabels.Add(this.myo2OrientationZ);
                //
                myoLabels.Add(this.myo3Address);
                myoLabels.Add(this.myo3Pose);
                myoLabels.Add(this.myo3GyroscopeX);
                myoLabels.Add(this.myo3GyroscopeY);
                myoLabels.Add(this.myo3GyroscopeZ);
                myoLabels.Add(this.myo3AccelerometerX);
                myoLabels.Add(this.myo3AccelerometerY);
                myoLabels.Add(this.myo3AccelerometerZ);
                myoLabels.Add(this.myo3OrientationW);
                myoLabels.Add(this.myo3OrientationX);
                myoLabels.Add(this.myo3OrientationY);
                myoLabels.Add(this.myo3OrientationZ);

                // Create handler
                myoHandler = new MyoHandler(
                    myoRateReductionLevel,
                    gyroscopeEnabled,
                    accelerometerEnabled,
                    orientationEnabled,
                    mqttHandler, 
                    oscHandler, 
                    this.Dispatcher, 
                    myoLabels);

                // Set source
                this.mqttSource.Content = Environment.MachineName;

                // Set topics
                this.mqttTopic1.Content = MyoHandler.MyoTopic1;
                this.mqttTopic2.Content = MyoHandler.MyoTopic2;
                this.mqttTopic3.Content = MyoHandler.MyoTopic3;
                this.mqttTopic4.Content = MyoHandler.MyoTopic4;

                // Set status
                myoGyroscopeStatus.Fill = gyroscopeEnabled ? Brushes.Green : Brushes.Red;
                myoAccelerometerStatus.Fill = accelerometerEnabled ? Brushes.Green : Brushes.Red;
                myoOrientationStatus.Fill = orientationEnabled ? Brushes.Green : Brushes.Red;
            }
        }

        #region Window handlers

        /// <summary>
        /// Startup program
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            if (mqttIsEnabled)
            {
                // Open MQTT 
                bool mqttIsConnected = mqttHandler.OpenMqtt();
                mqttStatusIndicator.Fill = mqttIsConnected ? Brushes.Green : Brushes.Red;
            }

            if (oscIsEnabled)
            {
                bool oscIsConnected = oscHandler.OpenOsc();
                oscStatusIndicator.Fill = oscIsConnected ? Brushes.Green : Brushes.Red;
            }

            if (myoIsEnabled)
            {
                // Open MYO
                bool myoIsConnected = myoHandler.OpenMyo();
                myoHubStatusIndicator.Fill = myoIsConnected ? Brushes.Green : Brushes.Red;
            }
        }

        /// <summary>
        /// Shutdown program
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Window_Closed(object sender, EventArgs e)
        {
            if (mqttIsEnabled)
            {
                // Close MQTT 
                mqttHandler.CloseMqtt();
                mqttStatusIndicator.Fill = Brushes.LightGray;
            }

            if (oscIsEnabled)
            {
                // Close OSC
                oscHandler.CloseOsc();
                oscStatusIndicator.Fill = Brushes.LightGray;
            }

            if (myoIsEnabled)
            {
                // Close Myo
                myoHandler.CloseMyo();
                myoHubStatusIndicator.Fill = Brushes.LightGray;
            }
        }

        #endregion

        private void mqttReconnect_Click(object sender, RoutedEventArgs e)
        {
            if (mqttIsEnabled)
            {
                // Close MQTT 
                mqttHandler.CloseMqtt();
                mqttStatusIndicator.Fill = Brushes.LightGray;

                // Update hostname from UI
                mqttHandler.BrokerHostname = mqttBroker.Text;

                // Open MQTT 
                bool mqttIsConnected = mqttHandler.OpenMqtt();
                mqttStatusIndicator.Fill = mqttIsConnected ? Brushes.Green : Brushes.Red;
            }
        }

        private void oscReconnect_Click(object sender, RoutedEventArgs e)
        {
            if (oscIsEnabled)
            {
                // Close OSC
                oscHandler.CloseOsc();
                oscStatusIndicator.Fill = Brushes.LightGray;

                // Update address and port from UI
                oscHandler.Address = oscAddress.Text;
                oscHandler.Port = oscPort.Text;

                // Open OSC
                bool oscIsConnected = oscHandler.OpenOsc();
                oscStatusIndicator.Fill = oscIsConnected ? Brushes.Green : Brushes.Red;
            }
        }

        private void myoHubReconnect_Click(object sender, RoutedEventArgs e)
        {
            if (myoIsEnabled)
            {
                // Close Myo
                myoHandler.CloseMyo();
                myoHubStatusIndicator.Fill = Brushes.LightGray;

                // Open MYO
                bool myoIsConnected = myoHandler.OpenMyo();
                myoHubStatusIndicator.Fill = myoIsConnected ? Brushes.Green : Brushes.Red;
            }
        }

        private void CheckBox_Checked(object sender, RoutedEventArgs e)
        {
            mqttHandler.SaveMessages = true;
        }

        private void CheckBox_Unchecked(object sender, RoutedEventArgs e)
        {
            mqttHandler.SaveMessages = false;
        }

        /// <summary>
        /// Configuration button click: opens Notepad to edit App.config
        /// </summary>
        private void Button_Click(object sender, RoutedEventArgs e)
        {
            ConfigurationFile.Edit();
        }
    }
}
