﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Threading;

using MyoSharp.Communication;
using MyoSharp.Device;
using MyoSharp.Exceptions;

using Newtonsoft.Json;

using eSkin4Common;

namespace eSkin4Myo
{
    public class MyoHandler
    {
        public const string TopicSourceTag = "[Source]";
        public const string TopicNameTag = "[Name]";
        public static string MyoTopic1 = "/[Name]/Pose";
        public static string MyoTopic2 = "/[Name]/Gyroscope";
        public static string MyoTopic3 = "/[Name]/Accelerometer";
        public static string MyoTopic4 = "/[Name]/Orientation";

        private bool myoIsConnected = false;
        private IChannelBridge myoChannelBridge;
        private IChannelDriver myoChannelDriver;
        private IChannel myoChannel;
        private IHub myoHub;
        private int rateReductionLevel;

        private long gyroscopeEventCount = 0;
        private long accelerometerEventCount = 0;
        private long orientationEventCount = 0;

        private bool gyroscopeEnabled = true;
        private bool accelerometerEnabled = true;
        private bool orientationEnabled = true;

        private MqttHandler mqttHandler = null;
        private OscHandler oscHandler = null;

        private Dispatcher uiDispatcher = null;
        private List<Label> uiLabels = null;
        private const int LabelsPerDevice = 12;

        /// <summary>
        /// Create a MYO handler.
        /// </summary>
        /// <param name="rateReductionLevel">Reduce sending rate by this factor.</param>
        /// <param name="gyroscopeEnabled">Enable gyroscope data messages</param>
        /// <param name="accelerometerEnabled">Enable accelerometer data messages</param>
        /// <param name="orientationEnabled">Enable orientation data messages</param>
        /// <param name="mqttHandler">MQTT handler to send messages to.</param>
        /// <param name="oscHandler">OSC handler to send messages to.</param>
        /// <param name="uiDispatcher">UI dispatcher for draw requests.</param>
        /// <param name="uiLabels">UI labels to populate with the received data.</param>
        public MyoHandler(
            // Device
            int rateReductionLevel,
            bool gyroscopeEnabled,
            bool accelerometerEnabled,
            bool orientationEnabled,
            // Messaging
            MqttHandler mqttHandler, 
            OscHandler oscHandler, 
            // UI
            Dispatcher uiDispatcher, 
            List<Label> uiLabels)
        {
            this.rateReductionLevel = (rateReductionLevel < 1) ? 1 : rateReductionLevel;
            this.gyroscopeEnabled = gyroscopeEnabled;
            this.accelerometerEnabled = accelerometerEnabled;
            this.orientationEnabled = orientationEnabled;
            this.mqttHandler = mqttHandler;
            this.oscHandler = oscHandler;
            this.uiDispatcher = uiDispatcher;
            this.uiLabels = uiLabels;
        }

        /// <summary>
        /// Get the connected state.
        /// </summary>
        public bool IsConnected
        {
            get { return myoIsConnected; }
        }

        #region MYO Handlers

        /// <summary>
        /// Connect MYO sensor
        /// </summary>
        public bool OpenMyo()
        {
            try
            {
                Console.WriteLine("Opening connection to MYO hub.");

                // create a hub that will manage Myo devices for us
                myoChannelBridge = ChannelBridge.Create();
                myoChannelDriver = ChannelDriver.Create(myoChannelBridge, MyoErrorHandlerDriver.Create(MyoErrorHandlerBridge.Create()));
                myoChannel = Channel.Create(myoChannelDriver);
                myoHub = Hub.Create(myoChannel);

                // Listen for when the Myo connects
                myoHub.MyoConnected += (sender, e) =>
                {
                    // Permanently unlock
                    e.Myo.Vibrate(VibrationType.Short);
                    e.Myo.Unlock(UnlockType.Hold);

                    // Hook up callbacks
                    e.Myo.PoseChanged += Myo_PoseChanged;
                    e.Myo.GyroscopeDataAcquired += Myo_GyroscopeDataAcquired;
                    e.Myo.AccelerometerDataAcquired += Myo_AccelerometerDataAcquired;
                    e.Myo.OrientationDataAcquired += Myo_OrientationDataAcquired;

                    string myoStatusText = string.Format("MYO {0} {1} has connected!", e.Myo.Handle, e.Myo.Name);
                    Console.WriteLine(myoStatusText);

                    // Update UI
                    uiDispatcher.Invoke((Action)(() =>
                    {
                        // Find empty slot and assign UI
                        int labelOffset =
                            (string.IsNullOrEmpty(uiLabels[0].Content.ToString())) ? 0 :
                            (string.IsNullOrEmpty(uiLabels[LabelsPerDevice].Content.ToString())) ? LabelsPerDevice :
                            (string.IsNullOrEmpty(uiLabels[2 * LabelsPerDevice].Content.ToString())) ? 2 * LabelsPerDevice : -1;
                        if (labelOffset != -1)
                        {
                            uiLabels[labelOffset + 0].Content = e.Myo.Name;
                        }
                    }));
                };

                // Listen for when the Myo disconnects
                myoHub.MyoDisconnected += (sender, e) =>
                {

                    e.Myo.PoseChanged -= Myo_PoseChanged;
                    e.Myo.GyroscopeDataAcquired -= Myo_GyroscopeDataAcquired;
                    e.Myo.AccelerometerDataAcquired -= Myo_AccelerometerDataAcquired;
                    e.Myo.OrientationDataAcquired -= Myo_OrientationDataAcquired;

                    string myoStatusText = string.Format("MYO {0} {1} has disconnected!", e.Myo.Handle, e.Myo.Name);
                    Console.WriteLine(myoStatusText);

                    // Update UI
                    uiDispatcher.Invoke((Action)(() =>
                    {
                        // Find slot and clear UI
                        int labelOffset = CalculateLabelOffset(e.Myo.Name);
                        if (labelOffset != -1)
                        {
                            for (int i = 0; i < LabelsPerDevice; i++)
                            {
                                uiLabels[labelOffset + i].Content = string.Empty;
                            }
                        }
                    }));
                };

                // start listening for Myo data
                myoChannel.StartListening();
                Console.WriteLine("Waiting for MYO connections.");
                myoIsConnected = true;
            }
            catch (Exception e)
            {
                Console.WriteLine("Error opening MYO sensors.");
                Console.WriteLine(e.Message);

                myoIsConnected = false;
            }

            return myoIsConnected;
        }

        /// <summary>
        /// Disconnect MYO sensor
        /// </summary>
        public void CloseMyo()
        {
            if (myoHub != null)
            {
                Console.WriteLine("Locking MYO sensors.");
                foreach (Myo myo in myoHub.Myos)
                {
                    myo.PoseChanged -= Myo_PoseChanged;
                    myo.GyroscopeDataAcquired -= Myo_GyroscopeDataAcquired;
                    myo.AccelerometerDataAcquired -= Myo_AccelerometerDataAcquired;
                    myo.OrientationDataAcquired -= Myo_OrientationDataAcquired;
                    myo.Lock();
                }
            }

            if (myoChannel != null)
            {
                Console.WriteLine("Closing connection to MYO sensors.");
                myoChannel.StopListening();
            }

            if (myoHub != null)
            {
                Console.WriteLine("Closing connection to MYO hub.");
                myoHub.Dispose();
                myoHub = null;
            }

            if (myoChannel != null)
            {
                myoChannel.Dispose();
                myoChannel = null;
            }

            myoIsConnected = false;
        }

        #region MYO Event Handlers

        private void Myo_PoseChanged(object sender, PoseEventArgs e)
        {
            // Create data record
            MyoPoseMessage myoData = new MyoPoseMessage();
            myoData.DeviceId = e.Myo.Handle.ToString();
            myoData.DeviceName = e.Myo.Name;
            myoData.Arm = e.Myo.Arm.ToString();
            myoData.Pose = e.Myo.Pose.ToString();

            // Update UI
            uiDispatcher.Invoke((Action)(() =>
            {
                int labelOffset = CalculateLabelOffset(myoData);
                if (labelOffset != -1)
                {
                    uiLabels[labelOffset + 1].Content = myoData.Pose;
                }
            }));

            // Send data
            string myoDataMessage = JsonConvert.SerializeObject(myoData);
            string topic = MyoTopic1.Replace(TopicSourceTag, myoData.Source).Replace(TopicNameTag, myoData.DeviceName);
            if (this.mqttHandler != null && this.mqttHandler.IsConnected)
            {
                this.mqttHandler.SendMessage(topic, myoDataMessage);
            }
            if (this.oscHandler != null && this.oscHandler.IsConnected)
            {
                this.oscHandler.SendMessage(topic, myoDataMessage);
            }
        }

        private void Myo_GyroscopeDataAcquired(object sender, GyroscopeDataEventArgs e)
        {
            if (!gyroscopeEnabled)
            {
                return;
            }

            // Rate reduction
            gyroscopeEventCount++;
            if (gyroscopeEventCount % rateReductionLevel == 0)
            {
                MyoValueMessage myoData = new MyoValueMessage();
                myoData.DeviceId = e.Myo.Handle.ToString();
                myoData.DeviceName = e.Myo.Name;
                myoData.Arm = e.Myo.Arm.ToString();
                myoData.W = null;
                myoData.X = Math.Round((double)e.Myo.Gyroscope.X, 3);
                myoData.Y = Math.Round((double)e.Myo.Gyroscope.Y, 3);
                myoData.Z = Math.Round((double)e.Myo.Gyroscope.Z, 3);

                // Update UI
                uiDispatcher.Invoke((Action)(() =>
                {
                    int labelOffset = CalculateLabelOffset(myoData);
                    if (labelOffset != -1)
                    {
                        uiLabels[labelOffset + 2].Content = myoData.X.ToString();
                        uiLabels[labelOffset + 3].Content = myoData.Y.ToString();
                        uiLabels[labelOffset + 4].Content = myoData.Z.ToString();
                    }
                }));

                // Send data
                string myoDataMessage = JsonConvert.SerializeObject(myoData);
                string topic = MyoTopic2.Replace(TopicSourceTag, myoData.Source).Replace(TopicNameTag, myoData.DeviceName);
                if (this.mqttHandler != null && this.mqttHandler.IsConnected)
                {
                    this.mqttHandler.SendMessage(topic, myoDataMessage);
                }
                if (this.oscHandler != null && this.oscHandler.IsConnected)
                {
                    this.oscHandler.SendMessage(topic, myoDataMessage);
                }
            }
        }

        private void Myo_AccelerometerDataAcquired(object sender, AccelerometerDataEventArgs e)
        {
            if (!accelerometerEnabled)
            {
                return;
            }

            // Rate reduction
            accelerometerEventCount++;
            if (accelerometerEventCount % rateReductionLevel == 0)
            {
                MyoValueMessage myoData = new MyoValueMessage();
                myoData.DeviceId = e.Myo.Handle.ToString();
                myoData.DeviceName = e.Myo.Name;
                myoData.Arm = e.Myo.Arm.ToString();
                myoData.W = null;
                myoData.X = Math.Round((double)e.Myo.Accelerometer.X, 3);
                myoData.Y = Math.Round((double)e.Myo.Accelerometer.Y, 3);
                myoData.Z = Math.Round((double)e.Myo.Accelerometer.Z, 3);

                // Update UI
                uiDispatcher.Invoke((Action)(() =>
                {
                    int labelOffset = CalculateLabelOffset(myoData);
                    if (labelOffset != -1)
                    {
                        uiLabels[labelOffset + 5].Content = myoData.X.ToString();
                        uiLabels[labelOffset + 6].Content = myoData.Y.ToString();
                        uiLabels[labelOffset + 7].Content = myoData.Z.ToString();
                    }
                }));

                // Send data
                string myoDataMessage = JsonConvert.SerializeObject(myoData);
                string topic = MyoTopic3.Replace(TopicSourceTag, myoData.Source).Replace(TopicNameTag, myoData.DeviceName);
                if (this.mqttHandler != null && this.mqttHandler.IsConnected)
                {
                    this.mqttHandler.SendMessage(topic, myoDataMessage);
                }
                if (this.oscHandler != null && this.oscHandler.IsConnected)
                {
                    this.oscHandler.SendMessage(topic, myoDataMessage);
                }
            }
        }

        private void Myo_OrientationDataAcquired(object sender, OrientationDataEventArgs e)
        {
            if (!orientationEnabled)
            {
                return;
            }

            // Rate reduction
            orientationEventCount++;
            if (orientationEventCount % rateReductionLevel == 0)
            {
                MyoValueMessage myoData = new MyoValueMessage();
                myoData.DeviceId = e.Myo.Handle.ToString();
                myoData.DeviceName = e.Myo.Name;
                myoData.Arm = e.Myo.Arm.ToString();
                myoData.W = Math.Round((double)e.Myo.Orientation.W, 3);
                myoData.X = Math.Round((double)e.Myo.Orientation.X, 3);
                myoData.Y = Math.Round((double)e.Myo.Orientation.Y, 3);
                myoData.Z = Math.Round((double)e.Myo.Orientation.Z, 3);

                // Update UI
                uiDispatcher.Invoke((Action)(() =>
                {
                    int labelOffset = CalculateLabelOffset(myoData);
                    if (labelOffset != -1)
                    {
                        uiLabels[labelOffset + 8].Content = myoData.W.ToString();
                        uiLabels[labelOffset + 9].Content = myoData.X.ToString();
                        uiLabels[labelOffset + 10].Content = myoData.Y.ToString();
                        uiLabels[labelOffset + 11].Content = myoData.Z.ToString();
                    }
                }));

                // Send data
                string myoDataMessage = JsonConvert.SerializeObject(myoData);
                string topic = MyoTopic4.Replace(TopicSourceTag, myoData.Source).Replace(TopicNameTag, myoData.DeviceName);
                if (this.mqttHandler != null && this.mqttHandler.IsConnected)
                {
                    this.mqttHandler.SendMessage(topic, myoDataMessage);
                }
                if (this.oscHandler != null && this.oscHandler.IsConnected)
                {
                    this.oscHandler.SendMessage(topic, myoDataMessage);
                }
            }
        }

        /// <summary>
        /// Find label offset in uiLabels list based on the deviceName in myoData.
        /// </summary>
        /// <param name="myoData">myoData object to get name from.</param>
        /// <returns>Offset into uiLabels list for name.</returns>
        private int CalculateLabelOffset(MyoBaseMessage myoData)
        {
            int labelOffset = CalculateLabelOffset(myoData.DeviceName);
            return labelOffset;
        }


        /// <summary>
        /// Find label offset in uiLabels list based on the deviceName .
        /// </summary>
        /// <param name="deviceName">Device name.</param>
        /// <returns>Offset into uiLabels list for name.</returns>
        private int CalculateLabelOffset(string deviceName)
        {
            int labelOffset =
                (uiLabels[0].Content.ToString().Equals(deviceName)) ? 0 :
                (uiLabels[LabelsPerDevice].Content.ToString().Equals(deviceName)) ? LabelsPerDevice :
                (uiLabels[2 * LabelsPerDevice].Content.ToString().Equals(deviceName)) ? 2 * LabelsPerDevice : -1;
            return labelOffset;
        }

        #endregion

        #endregion
    }
}
